
//***************/ SMC-S (SimTools Motor Control for Sabertooth 2x32) /************

#include <EEPROM.h>

int target1 = 0;             // "A" value from SIMTools       -2047 - 2047
int target2 = 0;             // "B" value from SIMTools       -2047 - 2047
volatile int feedback1 = 0;  // "C" value from Encoder        -2047 - 2047
volatile int feedback2 = 0;  // "D" value from Encoder        -2047 - 2047
int pwm1 = 0;                // "E" value from PID output     -2047 - 2047
int pwm2 = 0;                // "F" value from PID output     -2047 - 2047
int current1 = 0;            // "G" value from Sabertooth     -2047 - 2047
int current2 = 0;            // "H" value from Sabertooth     -2047 - 2047
int temp1;                   // "I" value from Sabertooth     0-4095
int temp2;                   // "J" value from Sabertooth     0-4095

int offsetM1_val;                // "M" value Processing <> EEprom     0-500
int offsetM2_val;                // "N" value Processing <> EEprom     0-500
int pwm_max_val = 1024;          // "P" value Processing <> EEprom     1024-4095
int kp_val = 255;                // "R" value Processing <> EEprom     0-4095
int ki_val = 0;                  // "S" value Processing <> EEprom     0-4095
int kd_val = 0;                  // "T" value Processing <> EEprom     0-4095
int speed_ramp_val = 1023;       // "V" value Processing <> EEprom     0-2047

byte rxByte = 0;                    // byte received from comm port
int bufferEnd = -1;                 // Rx Buffer end index
byte rxBuffer[5] = { 0 };           // 5 byte Rx Command Buffer
byte rxByte1 = 0;
int bufferEnd1 = -1;
byte rxBuffer1[5] = { 0 };
byte rxByte2 = 0;
int bufferEnd2 = -1;
byte rxBuffer2[8] = { 0 };
String inString2 = "";

byte txBuffer1[5] = { 0 };

boolean manual = false;

float kp, ki, kd;
float sampleTime = 0.008; // 8 msec
float outMax = 4095.0;    // PID output max
float lastErr1;
float lastErr2;
float iTerm1;
float iTerm2;

boolean target1Freeze = false;
boolean target2Freeze = false;
boolean pwmFreeze = true;

boolean timeTest = false;
boolean timeTest_old = false;
unsigned int loopCount = 0;

boolean channelSwitch;
boolean connectProc = false;      // connected to Processing via Bluetooth

char sab[20];

boolean encA1 = false;
boolean encA1_old = false;
boolean encA2 = false;
boolean encA2_old = false;
boolean forkM1;
boolean forkM2;
boolean new_SimState;
boolean old_SimState = false;

int zeroAdjStepM1 = 3;    // zero adjust step Motor 1
int zeroAdjStepM2 = 3;    // zero adjust step Motor 2

//*****************/ SETUP /********************************************

void setup()
{
  Serial.begin(500000);     // SimTools on Serial
  Serial1.begin(115200);    // Bluetooth on Serial 1
  Serial2.begin(38400);     // Sabertooth on Serial 2

  readEEprom();

  sprintf(sab, "R1:%d\n", speed_ramp_val);    // send SpeedRamping values to Sabertooth
  Serial2.write(sab);
  sprintf(sab, "R2:%d\n", speed_ramp_val);
  Serial2.write(sab);

  pinMode(4, INPUT);      // fork lightswitch M1
  pinMode(5, INPUT);      // fork lightswitch M2
  pinMode(6, OUTPUT);     // LED
  pinMode(7, OUTPUT);     // LED
  
  pinMode(A0, OUTPUT);     // LED
  pinMode(A1, OUTPUT);     // LED
  pinMode(A2, OUTPUT);     // LED
  pinMode(A3, OUTPUT);     // LED
  pinMode(A4, OUTPUT);     // LED
  pinMode(A5, OUTPUT);     // LED
  pinMode(A6, OUTPUT);     // LED
  pinMode(A7, OUTPUT);     // LED
  pinMode(A8, INPUT_PULLUP);     // low means button "Start zero adjust sequence" pressed
  pinMode(A9, INPUT);            // high means SIM switched on
  pinMode(A10, OUTPUT);    // Relay 1
  pinMode(A11, OUTPUT);    // Relay 2
  pinMode(A12, OUTPUT);    // Relay 3
  pinMode(A13, OUTPUT);    // Relay 4
  bitSet(PORTK, 2);     // A10 - Relay 1 off
  bitClear(PORTK, 3);   // A11 - Relay 2 on    Fans
  bitSet(PORTK, 4);     // A12 - Relay 3 off   PSU 1
  bitSet(PORTK, 5);     // A13 - Relay 4 off   PSU 2

  pinMode(2, INPUT_PULLUP);      // Quad Encoder inputs
  pinMode(3, INPUT_PULLUP);
  pinMode(20, INPUT_PULLUP);
  pinMode(21, INPUT_PULLUP);

  attachInterrupt(digitalPinToInterrupt(20), encoder1, CHANGE);    // INT1 D20
  attachInterrupt(digitalPinToInterrupt(2), encoder2, CHANGE);     // INT4 D2
  
  kp = float(kp_val) / 409.5;                      // set PID tunings
  ki = float(ki_val) / 409.5 * sampleTime;
  kd = float(kd_val) / 8191.0 / sampleTime;

  outMax = float(pwm_max_val >> 1);
}

//*****************/ MAIN LOOP /*********************************************

void loop()
{
  pollSerials();

  timeTest = micros() & bit(11);
  if (timeTest & ! timeTest_old)                             // runs once every 4 msec.
  {
    new_SimState = PINK & 1<<1;     // read A9   SIM switched on
    if ((new_SimState & ! old_SimState) | ! (PINK & 1<<0)) { zeroAdjust(); }   // pos. edge A9 or button A8 pressed: start zero adjust sequence
    else if ( ! new_SimState & old_SimState) { pauseSim(); }                   // neg. edge A9: shut down
    old_SimState = new_SimState;
    channelSwitch = ! channelSwitch;

    if (channelSwitch)                                       // runs once every 8 msec.
    {
      bitSet(PORTF, 0);                                                         // LED A0 on: "Channel 1 running"
      forkM1 = PING & 1<<5;                                                     // read D4
      if (forkM1) { bitSet(PORTF, 6); } else { bitClear(PORTF, 6); }            // LED A6
      if (zeroAdjStepM1 < 3) { seekZeroM1(); }                                  // Zero adjust sequence M1
      if (connectProc) { sendCommand1('C', feedback1 + 2047); }                 // send feedback value to processing
      pwm1 = pid1(feedback1, target1);                                          // PID calculation
      if (connectProc) { sendCommand1('E', pwm1 + 2047); }                      // send PWM value to processing
      if ( ! pwmFreeze) { sprintf(sab, "M1:%d\n", pwm1); Serial2.write(sab); }  // send PWM to sabertooth
      bitClear(PORTF, 0);                                                       // LED A0 off
    }
    else                                                      // runs once every 8 msec.
    {
      bitSet(PORTF, 1);                                                         // LED A1 on: "Channel 2 running"
      forkM2 = PINE & 1<<3;                                                     // read D5
      if (forkM2) { bitSet(PORTF, 7); } else { bitClear(PORTF, 7); }            // LED A7
      if (zeroAdjStepM2 < 3) { seekZeroM2(); }                                  // Zero adjust sequence M2
      if (connectProc) { sendCommand1('D', feedback2 + 2047); }                 // send feedback value to processing
      pwm2 = pid2(feedback2, target2);                                          // PID calculation
      if (connectProc) { sendCommand1('F', pwm2 + 2047); }                      // send PWM value to processing
      if ( ! pwmFreeze) { sprintf(sab, "M2:%d\n", pwm2); Serial2.write(sab); }  // send PWM to sabertooth
      bitClear(PORTF, 1);                                                       // LED A1 off
    }

    if (connectProc) { distributor(); }                      // run if connected to Processing
  }
  timeTest_old = timeTest;
}

//*****************/ action distributor /***************************************

void distributor()                 // 16bit timer , runs once every 4 msec. if connected to Processing
{
  if ((loopCount & 0x7) == 1)      // runs once every 32 msec.
  {
    Serial2.println("M1:getc");                 // "Current" value request to Sabertooth
  }
  if ((loopCount & 0x7) == 4)      // runs once every 32 msec.
  {
    Serial2.println("M2:getc");
  }
  if ((loopCount & 0xFFF) == 99)   // runs once every 16 sec.
  {
    Serial2.println("M1:gett");                 // "Temperature" value request to Sabertooth
    Serial2.println("M2:gett");
  }
  loopCount ++;
}

//*****************/ PID stuff /***************************************

int pid1(int input, int setpoint)   // PID calculation M1
{
  float inp = float(input);
  float error = float(setpoint) - inp;
  if (new_SimState)                             // if SIM switched on
  {
    iTerm1 += ki * error;
    if (iTerm1 > outMax) { iTerm1 = outMax; }
    else if (iTerm1 < -outMax) { iTerm1 = -outMax; }
  }
  else { iTerm1 = 0; }
  float dErr = error - lastErr1;
  lastErr1 = error;

  float out = (kp * error) + iTerm1 + (kd * dErr);
  if (out > outMax) { out = outMax; }
  else if (out < -outMax) { out = -outMax; }
  return int(out);
}

//-----------------------------------

int pid2(int input, int setpoint)   // PID calculation M2
{
  float inp = float(input);
  float error = float(setpoint) - inp;
  if (new_SimState)                           // if SIM switched on
  {
    iTerm2 += ki * error;
    if (iTerm2 > outMax) { iTerm2 = outMax; }
    else if (iTerm2 < -outMax) { iTerm2 = -outMax; }
  }
  else { iTerm2 = 0; }
  float dErr = error - lastErr2;
  lastErr2 = error;

  float out = (kp * error) + iTerm2 + (kd * dErr);
  if (out > outMax) { out = outMax; }
  else if (out < -outMax) { out = -outMax; }
  return int(out);
}

//*****************/ Zero adjust stuff /***************************************

void zeroAdjust()
{
  if (new_SimState)      // if SIM switched on
  {
    target1Freeze = true;
    target2Freeze = true;
    target1 = 0;
    target2 = 0;
    feedback1 = 0;
    feedback2 = 0;
    iTerm1 = 0;
    iTerm2 = 0;
    bitClear(PORTK, 4);     // A12 - Relay 3 (PSU 1) on
    bitClear(PORTK, 5);     // A13 - Relay 4 (PSU 2) on
    delay(1000);  
    pwmFreeze = false;
    zeroAdjStepM1 = 0;
    zeroAdjStepM2 = 0;
    bitSet(PORTF, 2);       // LED A2 on
    bitSet(PORTF, 3);       // LED A3 on
  }
}

//-----------------------------------

void seekZeroM1()            // Zero adjust sequence Motor 1
{
  if (zeroAdjStepM1 == 0)
  {
    if (forkM1) { zeroAdjStepM1 = 1; }
    else { target1 -= 3; }
  } 
  else if (zeroAdjStepM1 == 1)
  {
    if ( ! forkM1)
    {
      zeroAdjStepM1 = 2;
      feedback1 = 0;
      target1 = 0;
      iTerm1 = 0;
    }
    else { target1 += 3; }
  }
  else if (zeroAdjStepM1 == 2)
  {
    if (target1 > offsetM1_val)
    {
      zeroAdjStepM1 = 3;
      feedback1 = 0;
      target1 = 0;
      iTerm1 = 0;
      bitClear(PORTF, 2);     // LED A2 off
      target1Freeze = false;
    }
    else { target1 += 3; }
  }
}

//-----------------------------------

void seekZeroM2()            // Zero adjust sequence Motor 2
{
  if (zeroAdjStepM2 == 0)
  {
    if (forkM2) { zeroAdjStepM2 = 1; }
    else { target2 -= 3; }
  } 
  else if (zeroAdjStepM2 == 1)
  {
    if ( ! forkM2)
    {
      zeroAdjStepM2 = 2;
      feedback2 = 0;
      target2 = 0;
      iTerm2 = 0;
    }
    else { target2 += 3; }
  }
  else if (zeroAdjStepM2 == 2)
  {
    if (target2 > offsetM2_val)
    {
      zeroAdjStepM2 = 3;
      feedback2 = 0;
      target2 = 0;
      iTerm2 = 0;
      bitClear(PORTF, 3);     // LED A3 off
      target2Freeze = false;
    }
    else { target2 += 3; }
  }
}

//-----------------------------------

void pauseSim()
{
  pwmFreeze = true;
  bitSet(PORTK, 4);     // A12 - Relay 3 (PSU 1) off
  bitSet(PORTK, 5);     // A13 - Relay 4 (PSU 2) off
}

//*****************/ Quad Encoder stuff /**********************************

void encoder1()         // Interrupt - routine
{
  encA1 = PIND & 1<<1;     // D20  INT1
  if(PIND & 1<<0)          // D21  INT0
  {
    if(encA1 && ! encA1_old)
    {
      if (feedback1 > -2047) { feedback1 --; }
    } 
    else
    {
      if (feedback1 < 2047) { feedback1 ++; }
    }
  }
  encA1_old = encA1;
}

//-----------------------------------

void encoder2()         // Interrupt - routine
{
  encA2 = PINE & 1<<4;     // D2  INT4
  if(PINE & 1<<5)          // D3  INT5
  {
    if(encA2 && ! encA2_old)
    {
      if (feedback2 < 2047) { feedback2 ++; }
    } 
    else
    {
      if (feedback2 > -2047) { feedback2 --; }
    }
  }
  encA2_old = encA2;
}

//*****************/ SERIAL stuff /***************************************

void pollSerials()
{
  while (Serial.available())        // If anything comes in Serial (USB SimTools)
  {
    if (bufferEnd == -1)
    {
      rxByte = Serial.read();
      if (rxByte != '[') { bufferEnd = -1; }
      else { bufferEnd = 0; }
    }
    else
    {
      rxByte = Serial.read();
      rxBuffer[bufferEnd] = rxByte;
      bufferEnd++;
      if (bufferEnd > 3)
      {
        if (rxBuffer[3] == ']') { parseCommand(); }
        bufferEnd = -1;
      }
    }
  }
  while (Serial1.available())        // If anything comes in Serial 1 (Processing)
  {
    if (bufferEnd1 == -1)
    {
      rxByte1 = Serial1.read();
      if (rxByte1 != '[') { bufferEnd1 = -1; }
      else { bufferEnd1 = 0; }
    }
    else
    {
      rxByte1 = Serial1.read();
      rxBuffer1[bufferEnd1] = rxByte1;
      bufferEnd1++;
      if (bufferEnd1 > 3)
      {
        if (rxBuffer1[3] == ']') { parseCommand1(); }
        bufferEnd1 = -1;
      }
    }
  }
  while (Serial2.available())        // If anything comes in Serial 2 (Sabertooth)
  {
    if (bufferEnd2 == -1)
    {
      rxByte2 = Serial2.read();
      if (rxByte2 != 'M') { bufferEnd2 = -1; }
      else { bufferEnd2 = 0; }
    }
    else
    {
      rxByte2 = Serial2.read();
      if (rxByte2 != 13)            // Carriage return ?
      {
        rxBuffer2[bufferEnd2] = rxByte2;
        bufferEnd2++;
      }
      else
      {
        rxBuffer2[bufferEnd2] = 13;
        parseCommand2();
        bufferEnd2 = -1;
      }
    }
  }
}
//-----------------------

void parseCommand()       // from serial USB (SimTools)
{
  int value =  (rxBuffer[1] << 8) + rxBuffer[2];
  switch (rxBuffer[0])
  {
    case 'A':                                      // <Axis1> from SimTools
      if ( ! manual && ! target1Freeze) { target1 = value - 2047; }
      sendCommand1('A', value);
      break;
    case 'B':                                      // <Axis2> from SimTools
      if ( ! manual && ! target2Freeze) { target2 = value - 2047; }
      sendCommand1('B', value);
//      break;
//    case 'S': zeroAdjust();    break;              // "Startup" from SimTools
//    case 'T': zeroAdjust();                        // "Shutdown" from SimTools
  }
}

//-----------------------

void parseCommand1()        // from serial 1 bluetooth (Processing)
{
//  Serial.println(rxBuffer1[0]);
  int value1 = (rxBuffer1[1] << 8) + rxBuffer1[2];
  switch (rxBuffer1[0])
  {
    case 'A': if (manual && ! target1Freeze) { target1 = value1 - 2047; }  break;
    case 'B': if (manual && ! target2Freeze) { target2 = value1 - 2047; }  break;
    case 'M': offsetM1_val = value1;                                       break;
    case 'N': offsetM2_val = value1;                                       break;
    case 'P': pwm_max_val = value1; outMax = float(value1 >> 1);           break;
    case 'R': kp_val = value1; kp = float(kp_val) / 409.5;                 break;     // set P-value
    case 'S': ki_val = value1; ki = float(ki_val) / 409.5 * sampleTime;    break;     // set I-value
    case 'T': kd_val = value1; kd = float(kd_val) / 8191.0 / sampleTime;   break;     // set D-value
    case 'V':                          // send Speed Ramping values to Sabertooth
      speed_ramp_val = value1;
      sprintf(sab, "R1:%d\n", value1);
      Serial2.write(sab);
      sprintf(sab, "R2:%d\n", value1);
      Serial2.write(sab);
      break;
    case 'X': if (manual && ! target1Freeze && ! target2Freeze)      // values 'A' + 'B' for both channels
              {
                target1 = value1 - 2047;
                target2 = target1;
              }
              break;
    case 'c':                     // connect to Processing
      connectProc = true;
      sendCommand1('c', 0);
      break;
    case 'd':                     // disconnect from Processing
      connectProc = false;
      target1 = 0;
      target2 = 0;
      manual = false;
      sendCommand1('d', 0);
      break;
    case 'm': manual = true;        break;       // mode 'manual'
    case 's': manual = false;       break;       // mode 'SimTools'
    case 'w': writeEEprom();        break;       // save values to EEprom
    case 'z': zeroAdjust();         break;       // start zero adjust sequence
    
    case 'r':                       // request from Processing
      switch (rxBuffer1[2])
      {
        case 'M': sendCommand1('M', offsetM1_val);    break;
        case 'N': sendCommand1('N', offsetM2_val);    break;
        case 'P': sendCommand1('P', pwm_max_val);     break;
        case 'R': sendCommand1('R', kp_val);          break;
        case 'S': sendCommand1('S', ki_val);          break;
        case 'T': sendCommand1('T', kd_val);          break;
        case 'V': sendCommand1('V', speed_ramp_val);
      }
  }
}

//-----------------------

void parseCommand2()       // from serial2 (sabertooth)
{
  String inString = "";
  int inVal;
  int n = 3;
  while (rxBuffer2[n] != 13)
  {
    if (isDigit(rxBuffer2[n])) { inString += (char)rxBuffer2[n]; }
    n++;
  }
  inVal = inString.toInt();
  if (rxBuffer2[3] == '-') { inVal = -inVal; }

  if ((rxBuffer2[0] == '1') & (rxBuffer2[2] == 'C'))
  {
    current1 = inVal * 5;
    if (current1 > 2047) { current1 = 2047; }
    if (current1 < -2047) { current1 = -2047; }
    sendCommand1('G', current1 + 2047);
  }
  else if ((rxBuffer2[0] == '2') & (rxBuffer2[2] == 'C'))
  {
    current2 = inVal * 5;
    if (current2 > 2047) { current2 = 2047; }
    if (current2 < -2047) { current2 = -2047; }
    sendCommand1('H', current2 + 2047);  }
  else if ((rxBuffer2[0] == '1') & (rxBuffer2[2] == 'T'))
  {
    temp1 = inVal * 30;
    sendCommand1('I', temp1);
  }
  else if ((rxBuffer2[0] == '2') & (rxBuffer2[2] == 'T'))
  {
    temp2 = inVal * 30;
    sendCommand1('J', temp2);
  }
}

//-----------------------

void sendCommand1(char sendComm, int sendVal)       // to serial 1 Processing
{
  txBuffer1[0] = '[';
  txBuffer1[1] = byte(sendComm);
  txBuffer1[2] = highByte(sendVal);
  txBuffer1[3] = lowByte(sendVal);
  txBuffer1[4] = ']';
  Serial1.write(txBuffer1, 5);
}

//******************/ EEPROM stuff /*****************************

void readEEprom()
{
  int EEcheck;
  EEPROM.get(0, EEcheck);
  if (EEcheck != 1234) { writeEEprom(); }     // if EEprom empty write default values
  EEPROM.get(4, offsetM1_val);
  EEPROM.get(6, offsetM2_val);
  EEPROM.get(8, pwm_max_val);
  EEPROM.get(12, kp_val);
  EEPROM.get(14, ki_val);
  EEPROM.get(16, kd_val);
  EEPROM.get(20, speed_ramp_val);
}

void writeEEprom()
{
  EEPROM.put(0, 1234);
  EEPROM.put(4, offsetM1_val);
  EEPROM.put(6, offsetM2_val);
  EEPROM.put(8, pwm_max_val);
  EEPROM.put(12, kp_val);
  EEPROM.put(14, ki_val);
  EEPROM.put(16, kd_val);
  EEPROM.put(20, speed_ramp_val);
}

//*********************************************





